/*--------------------------------------------*/
/*   Sample code(2) of using SSI32 on MPI     */
/*   (count hexa-decimal characters appeared  */
/*    in hexa-decimal representations of      */
/*    successive 0x400000000 random numbers)  */
/*--------------------------------------------*/

/*  Notes: 1. The result does not depend on mpisize.
           2. On a single PC(Windows XP, x64, icl) about
              one hour is needed to get the result. */

#define useMPI    // if not use MPI, remove this line
#define Use_icc   // if your compiler is icc, use this line

unsigned long long int TotalRandomNumbers=0x400000000;

#include <stdio.h>
#include <time.h>
#ifdef useMPI
#include <mpi.h>   /*MPI*/
#endif

#include "SSI32rand.c"

/*******************/

#ifdef useMPI
int main(int argc, char * argv[])
#else
int main()
#endif
    {
    #ifdef useMPI
    MPI_Status mpistat;   /*MPI*/
    #endif
    int  mpierr=0;
    int  mpisize=1, myrank=0;
    unsigned long long int RandomNumbersInMyrank;
    unsigned long long int NoOfHexCharsULLI[16];
    unsigned int NoOfHexCharsUI[32];
    unsigned long long int ulli_i;
    unsigned int SSI32RN;
    int  i, RecvNo;
    time_t  startTime, endTime;
    
    #ifdef useMPI
    MPI_Init(&argc, &argv);                   /*MPI*/
    MPI_Comm_size(MPI_COMM_WORLD, &mpisize);  /*MPI*/
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);   /*MPI*/
    #endif
    if (myrank==0)
        {
        printf("mpisize = %d\n", mpisize);
        printf("Total SSI32_random_numbers are 0x%x",
                                 (int)(TotalRandomNumbers>>32));
        printf("%08x\n\n",(int)(TotalRandomNumbers & 0xffffffff));
        }

    for (i=0; i<16; i++) {NoOfHexCharsULLI[i]=0;}
    RandomNumbersInMyrank=(TotalRandomNumbers/mpisize);
    if (myrank==(mpisize-1))
        {RandomNumbersInMyrank += (TotalRandomNumbers%mpisize);}

    time (&startTime);
    SSI32IniKnm(TotalRandomNumbers/mpisize, (unsigned long long int)myrank);
    for (ulli_i=0; ulli_i<RandomNumbersInMyrank; ulli_i++)
        {
        SSI32RN=SSI32KBin32();
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[(SSI32RN & 0xf)]++;
        SSI32RN >>= 4;
        NoOfHexCharsULLI[SSI32RN]++;
        }

    if (myrank==0)
        {
        for (RecvNo=1; RecvNo<mpisize; RecvNo++)
            {
            #ifdef useMPI
            mpierr=MPI_Recv(NoOfHexCharsUI, 32, MPI_INT,
                            RecvNo, 1234, MPI_COMM_WORLD, &mpistat);  /*MPI*/
            #endif
            if (mpierr) {printf("mpi recv err : myrank=%d\n", RecvNo);}
            for (i=0; i<16; i++)
                {
                ulli_i = NoOfHexCharsUI[i*2+1];
                ulli_i <<= 32;
                ulli_i |= NoOfHexCharsUI[i*2];
                NoOfHexCharsULLI[i] += ulli_i;
                }
            }
        }
      else
        {
        for (i=0; i<16; i++)
            {
            NoOfHexCharsUI[i*2] = (NoOfHexCharsULLI[i] & 0xffffffff);
            NoOfHexCharsUI[i*2+1] = (NoOfHexCharsULLI[i] >> 32);
            }
        #ifdef useMPI
        mpierr=MPI_Send(NoOfHexCharsUI, 32, MPI_INT,
                        0, 1234, MPI_COMM_WORLD);                    /*MPI*/
        #endif
        if (mpierr) {printf("mpi send err : %d\n", RecvNo);}
        }
    time (&endTime);
    #ifdef useMPI
    MPI_Finalize();   /*MPI*/
    #endif
 
    if (myrank==0)
        {
        ulli_i=0;
        for (i=0; i<16; i++)
            {
            printf("[%x] 0x%x", i, (int)(NoOfHexCharsULLI[i] >> 32));
            printf("%08x\n", (int)(NoOfHexCharsULLI[i] & 0xffffffff));
            ulli_i+=NoOfHexCharsULLI[i];
            }
            /* For TotalRandomNumbers=0x400000000, the result is
              [0] 0x200017e99    [1] 0x20000a512    [2] 0x1fffbf9a2
              [3] 0x20001d0e3    [4] 0x20000df65    [5] 0x1fffe51b2
              [6] 0x20000dda3    [7] 0x1ffff12dd    [8] 0x200000dac
              [9] 0x20001879d    [a] 0x20001c8c4    [b] 0x1ffff15e5
              [c] 0x1fffe4ebe    [d] 0x20000763c    [e] 0x1ffffae15
              [f] 0x200000938  */
        printf ("\nTime = %d seconds %6.2f minutes\n",
                (int)(endTime-startTime), ((int)(endTime-startTime))/60.0);
        if (ulli_i != (TotalRandomNumbers*8))
            {printf("\nError: total Hex chars != TotalRandomNumbers*8\n");}
        }

    return(0);
    }

